/**********************************************************************
Copyright (c) 2008 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.datanucleus.store.odf;

import java.io.File;
import java.util.Map;

import javax.transaction.xa.XAResource;

import org.datanucleus.OMFContext;
import org.datanucleus.exceptions.NucleusException;
import org.datanucleus.store.connection.AbstractConnectionFactory;
import org.datanucleus.store.connection.AbstractManagedConnection;
import org.datanucleus.store.connection.ManagedConnection;
import org.datanucleus.store.connection.ManagedConnectionResourceListener;
import org.openoffice.odf.doc.OdfDocument;
import org.openoffice.odf.doc.OdfFileDom;
import org.openoffice.odf.doc.OdfSpreadsheetDocument;
import org.openoffice.odf.doc.element.office.OdfAutomaticStyles;
import org.openoffice.odf.doc.element.style.OdfStyle;
import org.openoffice.odf.doc.element.style.OdfTableCellProperties;
import org.openoffice.odf.doc.element.style.OdfTableRowProperties;
import org.openoffice.odf.doc.element.table.OdfTable;
import org.openoffice.odf.dom.OdfNamespace;
import org.openoffice.odf.dom.style.OdfStyleFamily;
import org.w3c.dom.Element;

/**
 * Implementation of a ConnectionFactory for ODF.
 */
public class ConnectionFactoryImpl extends AbstractConnectionFactory
{
    /**
     * Constructor.
     * @param omfContext The OMF context
     * @param resourceType Type of resource (tx, nontx)
     */
    public ConnectionFactoryImpl(OMFContext omfContext, String resourceType)
    {
        super(omfContext, resourceType);
    }

    public ManagedConnection createManagedConnection(Object poolKey, Map transactionOptions)
    {
        return new ManagedConnectionImpl(omfContext, transactionOptions);
    }

    public class ManagedConnectionImpl extends AbstractManagedConnection
    {
        OMFContext omf;
        Map options;

        /** The ODF file. */
        File file;

        public ManagedConnectionImpl(OMFContext omf, Map options)
        {
            this.omf = omf;
            this.options = options;
        }

        public void close()
        {
            if (conn == null)
            {
                return;
            }

            for (int i=0; i<listeners.size(); i++)
            {
                ((ManagedConnectionResourceListener)listeners.get(i)).managedConnectionPreClose();
            }
            try
            {
                //close
                //this is a kind of commit.... do it somewhere else
                ((OdfDocument)conn).save(file);
                conn = null;
            }
            catch (Exception e)
            {
                throw new NucleusException(e.getMessage(),e);
            }
            finally
            {
                for (int i=0; i<listeners.size(); i++)
                {
                    ((ManagedConnectionResourceListener)listeners.get(i)).managedConnectionPostClose();
                }
            }
        }

        public Object getConnection()
        {
            if (conn == null)
            {
                // "openoffice:file:{filename}"
                String url = omf.getStoreManager().getConnectionURL();
                if (url == null)
                {
                    throw new NucleusException("you haven't specified persistence property 'datanucleus.ConnectionURL' (or alias)");
                }
                if (!url.startsWith("odf:"))
                {
                    throw new NucleusException("invalid URL: "+url);
                }

                // Split the URL into filename
                String str = url.substring("odf:".length()); // Omit "odf:"
                if (str.indexOf("file:") != 0)
                {
                    throw new NucleusException("invalid URL: "+url);
                }

                try
                {
                    String filename = str.substring("file:".length()); // Omit "file:"
                    file = new File(filename);
                    if (!file.exists())
                    {
                        // ODF spreadsheet doesn't exist, so create
                        OdfDocument doc = OdfSpreadsheetDocument.createSpreadsheetDocument();

                        // Remove the default table added in construction
                        OdfFileDom contentDoc = doc.getContentDom();
                        Element spreadsheetElement = (Element)contentDoc.getElementsByTagNameNS(
                            OdfNamespace.OFFICE.getUri(), "spreadsheet").item(0);
                        OdfTable table = (OdfTable) (spreadsheetElement.getElementsByTagNameNS(
                            OdfNamespace.TABLE.getUri(), "table").item(0));
                        table.getParentNode().removeChild(table);

                        doc.save(file);
                    }

                    conn = OdfDocument.loadDocument(file);

                    // Make sure we have all required styles
                    OdfFileDom contentDoc = ((OdfDocument)conn).getContentDom();
                    OdfAutomaticStyles styles = contentDoc.getOrCreateAutomaticStyles();

                    // ColumnHeader colouring
                    OdfStyle style = styles.getStyle("DN_Headers", OdfStyleFamily.TableRow);
                    if (style == null)
                    {
                        style = new OdfStyle(contentDoc);
                        style.setName("DN_Headers");
                        style.setFamily(OdfStyleFamily.TableRow);
                        style.setProperty(OdfTableRowProperties.BackgroundColor, "#74a3db");
                        styles.appendChild(style);
                    }

                    // Primary-Key colouring
                    style = styles.getStyle("DN_PK", OdfStyleFamily.TableCell);
                    if (style == null)
                    {
                        style = new OdfStyle(contentDoc);
                        style.setName("DN_PK");
                        style.setFamily(OdfStyleFamily.TableCell);
                        style.setProperty(OdfTableCellProperties.BackgroundColor, "#c2d9e0");
                        styles.appendChild(style);
                    }

                    // Relation colouring
                    style = styles.getStyle("DN_Relation", OdfStyleFamily.TableCell);
                    if (style == null)
                    {
                        style = new OdfStyle(contentDoc);
                        style.setName("DN_Relation");
                        style.setFamily(OdfStyleFamily.TableCell);
                        style.setProperty(OdfTableCellProperties.BackgroundColor, "#DDDDDD");
                        styles.appendChild(style);
                    }
                }
                catch (Exception e)
                {
                    throw new NucleusException(e.getMessage(),e);
                }
            }
            return conn;
        }

        public XAResource getXAResource()
        {
            return null;
        }
    }
}