/**********************************************************************
Copyright (c) 2009 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.datanucleus.store.odf;

import java.util.ArrayList;
import java.util.List;

import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.FetchPlan;
import org.datanucleus.exceptions.NucleusDataStoreException;
import org.datanucleus.exceptions.NucleusException;
import org.datanucleus.identity.OID;
import org.datanucleus.identity.OIDFactory;
import org.datanucleus.metadata.AbstractClassMetaData;
import org.datanucleus.metadata.AbstractMemberMetaData;
import org.datanucleus.metadata.ColumnMetaData;
import org.datanucleus.metadata.IdentityMetaData;
import org.datanucleus.metadata.IdentityType;
import org.datanucleus.metadata.VersionMetaData;
import org.datanucleus.store.ExecutionContext;
import org.datanucleus.store.FieldValues2;
import org.datanucleus.store.ObjectProvider;
import org.datanucleus.store.Type;
import org.datanucleus.store.connection.ManagedConnection;
import org.datanucleus.store.odf.fieldmanager.FetchFieldManager;
import org.openoffice.odf.doc.OdfDocument;
import org.openoffice.odf.doc.OdfFileDom;
import org.openoffice.odf.doc.element.table.OdfTable;
import org.openoffice.odf.doc.element.table.OdfTableCell;
import org.openoffice.odf.doc.element.table.OdfTableRow;
import org.openoffice.odf.dom.OdfNamespace;
import org.openoffice.odf.dom.type.office.OdfValueType;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Utilities to assist in persistence to ODF spreadsheets.
 */
public class ODFUtils
{
    /** String value to use to represent null string. */
    public final static String STRING_NULL_VALUE = "#NULL";

    /**
     * Convenience method to return the required sheet name for the specified class.
     * @param cmd Metadata for the class
     * @return The name of the sheet (table)
     */
    public static String getSheetNameForClass(AbstractClassMetaData cmd)
    {
        String sheetName = cmd.getValueForExtension("sheet");
        if (sheetName == null)
        {
            if (cmd.getTable() != null)
            {
                sheetName = cmd.getTable();
            }
            else
            {
                sheetName = cmd.getName();
            }
        }
        return sheetName;
    }

    /**
     * Convenience method to find the sheet (table) node appropriate for the specified class.
     * @param cmd Metadata for the class
     * @param contentDoc The content document
     * @return The table node (or null of not present)
     */
    public static OdfTable getODFTableForClass(AbstractClassMetaData cmd, OdfFileDom contentDoc)
    {
        OdfTable table = null;
        String sheetNameForClass = getSheetNameForClass(cmd);
        Element spreadsheetElement = (Element)contentDoc.getElementsByTagNameNS(
            OdfNamespace.OFFICE.getUri(), "spreadsheet").item(0);
        NodeList tables = spreadsheetElement.getElementsByTagNameNS(OdfNamespace.TABLE.getUri(), "table");
        for (int i=0;i<tables.getLength();i++)
        {
            Node tableNode = tables.item(i);
            NamedNodeMap attrMap = tableNode.getAttributes();
            for (int j=0;j<attrMap.getLength();j++)
            {
                Node attr = attrMap.item(j);
                if (attr.getNodeName().equals("table:name") && attr.getNodeValue().equals(sheetNameForClass))
                {
                    return (OdfTable)tableNode;
                }
            }
        }
        return table;
    }

    /**
     * Convenience method to find the row of an object in the provided sheet.
     * For application-identity does a search for a row with the specified PK field values.
     * For datastore-identity does a search for the row with the datastore column having the specified value
     * @param sm StateManager for the object
     * @param contentDoc The content document
     * @return The row (or null if not found)
     */
    public static OdfTableRow getTableRowForObjectInSheet(ObjectProvider sm, OdfFileDom contentDoc)
    {
        final AbstractClassMetaData cmd = sm.getClassMetaData();
        OdfTable table = getODFTableForClass(sm.getClassMetaData(), contentDoc);
        if (table == null)
        {
            return null;
        }

        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            int[] pkFieldNumbers = cmd.getPKMemberPositions();
            Object[] pkFieldValues = new Object[pkFieldNumbers.length];
            for (int i=0;i<pkFieldNumbers.length;i++)
            {
                pkFieldValues[i] = sm.provideField(pkFieldNumbers[i]);
            }

            NodeList rowNodes = table.getChildNodes();
            for (int i=0;i<rowNodes.getLength();i++)
            {
                Node rowNode = rowNodes.item(i);
                if (rowNode instanceof OdfTableRow)
                {
                    NodeList cellNodes = rowNode.getChildNodes();
                    boolean isRow = true;

                    for (int j=0;j<pkFieldNumbers.length;j++)
                    {
                        int index = getColumnIndexForFieldOfClass(cmd, pkFieldNumbers[j]);
                        OdfTableCell cellNode = (OdfTableCell)cellNodes.item(index);
                        Object pkVal = pkFieldValues[j];
                        if (!doesCellMatchValue(cellNode, pkVal))
                        {
                            isRow = false;
                            break;
                        }
                    }
                    if (isRow)
                    {
                        return (OdfTableRow)rowNode;
                    }
                }
            }
        }
        else if (cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            OID oid = (OID)sm.getInternalObjectId();
            Object key = oid.getKeyValue();
            int index = getColumnIndexForFieldOfClass(cmd, -1);
            NodeList rowNodes = table.getChildNodes();
            for (int i=0;i<rowNodes.getLength();i++)
            {
                Node rowNode = rowNodes.item(i);
                if (rowNode instanceof OdfTableRow)
                {
                    NodeList cellNodes = rowNode.getChildNodes();
                    OdfTableCell cell = (OdfTableCell)cellNodes.item(index);
                    if (doesCellMatchValue(cell, key))
                    {
                        return (OdfTableRow)rowNode;
                    }
                }
            }
        }

        return null;
    }

    /**
     * Convenience method to return if the supplied cell matches the passed value.
     * Compares the type and value.
     * @param cell The cell
     * @param value The value
     * @return Whether they match
     */
    private static boolean doesCellMatchValue(OdfTableCell cell, Object value)
    {
        if (cell == null)
        {
            return false;
        }

        if (value instanceof Long && cell.getValueType() == OdfValueType.FLOAT && 
            cell.getValue().longValue() == ((Long)value).longValue())
        {
            return true;
        }
        else if (value instanceof Integer && cell.getValueType() == OdfValueType.FLOAT &&
            cell.getValue().intValue() == ((Integer)value).intValue())
        {
            return true;
        }
        else if (value instanceof Short && cell.getValueType() == OdfValueType.FLOAT &&
            cell.getValue().shortValue() == ((Short)value).shortValue())
        {
            return true;
        }
        else if (value instanceof String && cell.getValueType() == OdfValueType.STRING && 
            cell.getStringValue().equals((String)value))
        {
            return true;
        }
        // TODO Cater for other types
        return false;
    }

    /**
     * Convenience method to get the index number where a field of a class is persisted.
     * Uses the extension tag "index" and if not provided uses "column" expecting an integer value.
     * The field number is the absolute number (0 or higher); a value of -1 implies surrogate identity column,
     * and -2 implies surrogate version column
     * @param acmd MetaData for the class
     * @param inputFieldNumber Absolute field number that we are interested in (-1 = datastore-id, -2=version)
     */
    public static int getColumnIndexForFieldOfClass(AbstractClassMetaData acmd, int inputFieldNumber)
    {
        int fieldNumber = inputFieldNumber;
        if (inputFieldNumber == -1)
        {
            // Datastore-identity, so allocate next column after normal fields
            fieldNumber = acmd.getNoOfManagedMembers() + acmd.getNoOfInheritedManagedMembers();
        }
        else if (inputFieldNumber == -2)
        {
            // Version, so allocate next column after normal fields (and optionally datastore-identity)
            if (acmd.getIdentityType() == IdentityType.DATASTORE)
            {
                fieldNumber = acmd.getNoOfManagedMembers() + acmd.getNoOfInheritedManagedMembers() + 1;
            }
            else
            {
                fieldNumber = acmd.getNoOfManagedMembers() + acmd.getNoOfInheritedManagedMembers();
            }
        }

        if (inputFieldNumber >= 0)
        {
            // Field of the class
            AbstractMemberMetaData ammd = acmd.getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
            String index = ammd.getValueForExtension("index");
            if (index == null)
            {
                ColumnMetaData[] colmds = ammd.getColumnMetaData();
                if (colmds != null && colmds.length > 0)
                {
                    String colName = colmds[0].getName();
                    try
                    {
                        return Integer.valueOf(colName).intValue();
                    }
                    catch (NumberFormatException nfe)
                    {
                        return (int)fieldNumber;
                    }
                }
                else
                {
                    return (int)fieldNumber;
                }
            }
            else
            {
                try
                {
                    return Integer.valueOf(index).intValue();
                }
                catch (NumberFormatException nfe)
                {
                    return (int)fieldNumber;
                }
            }
        }
        else if (inputFieldNumber == -1)
        {
            // Surrogate datastore identity column
            IdentityMetaData imd = acmd.getIdentityMetaData();
            if (imd != null)
            {
                String index = imd.getValueForExtension("index");
                if (index == null)
                {
                    if (imd.getColumnMetaData() != null && imd.getColumnMetaData().length > 0)
                    {
                        String colName = imd.getColumnMetaData()[0].getName();
                        try
                        {
                            return Integer.valueOf(colName).intValue();
                        }
                        catch (NumberFormatException nfe)
                        {
                            return (int)fieldNumber;
                        }
                    }
                    else
                    {
                        return (int)fieldNumber;
                    }
                }
                else
                {
                    try
                    {
                        return Integer.valueOf(index).intValue();
                    }
                    catch (NumberFormatException nfe)
                    {
                        return (int)fieldNumber;
                    }
                }
            }
            return -1;
        }
        else if (inputFieldNumber == -2)
        {
            // Surrogate version column
            VersionMetaData vmd = acmd.getVersionMetaData();
            if (vmd != null)
            {
                String index = vmd.getValueForExtension("index");
                if (index == null)
                {
                    if (vmd.getColumnMetaData() != null && vmd.getColumnMetaData().length > 0)
                    {
                        String colName = vmd.getColumnMetaData()[0].getName();
                        try
                        {
                            return Integer.valueOf(colName).intValue();
                        }
                        catch (NumberFormatException nfe)
                        {
                            return (int)fieldNumber;
                        }
                    }
                    else
                    {
                        return (int)fieldNumber;
                    }
                }
                else
                {
                    try
                    {
                        return Integer.valueOf(index).intValue();
                    }
                    catch (NumberFormatException nfe)
                    {
                        return (int)fieldNumber;
                    }
                }
            }
            return -1;
        }
        else
        {
            throw new NucleusException("Unsupported field number " + fieldNumber);
        }
    }

    /**
     * Convenience method to get all objects of the candidate type (and optional subclasses) from the 
     * specified workbook connection.
     * @param ec execution context
     * @param mconn Managed Connection
     * @param candidateClass Candidate
     * @param subclasses Include subclasses?
     * @param ignoreCache Whether to ignore the cache
     * @return List of objects of the candidate type
     */
    public static List getObjectsOfCandidateType(ExecutionContext ec, ManagedConnection mconn, 
            Class candidateClass, boolean subclasses, boolean ignoreCache)
    {
        OdfDocument doc = (OdfDocument)mconn.getConnection();
        List results = null;
        try
        {
            OdfFileDom contentDoc = doc.getContentDom();
            ClassLoaderResolver clr = ec.getClassLoaderResolver();
            final AbstractClassMetaData acmd = ec.getMetaDataManager().getMetaDataForClass(candidateClass, clr);
            results = getObjectsOfCandidateType(ec, contentDoc, acmd, ignoreCache);
            if (subclasses)
            {
                // Add on any subclass objects
                String[] subclassNames = ec.getMetaDataManager().getSubclassesForClass(candidateClass.getName(), true);
                if (subclassNames != null)
                {
                    for (int i=0;i<subclassNames.length;i++)
                    {
                        AbstractClassMetaData cmd = ec.getMetaDataManager().getMetaDataForClass(subclassNames[i], clr);
                        results.addAll(getObjectsOfCandidateType(ec, contentDoc, cmd, ignoreCache));
                    }
                }
            }
        }
        catch (Exception e)
        {
            throw new NucleusDataStoreException("Error extracting results from ODF document", e);
        }
        return results;
    }

    /**
     * Method to extract the actual objects of a particular class from the Excel spreadsheet.
     * @param ec execution context
     * @param wb Workbook
     * @param acmd MetaData for the class
     * @param ignoreCache Whether to ignore the cache
     * @return List of objects (connected to StateManagers as required)
     */
    private static List getObjectsOfCandidateType(ExecutionContext ec, OdfFileDom contentDoc,
            final AbstractClassMetaData acmd, boolean ignoreCache)
    {
        List results = new ArrayList();

        ClassLoaderResolver clr = ec.getClassLoaderResolver();
        final OdfTable table = ODFUtils.getODFTableForClass(acmd, contentDoc);
        if (table != null)
        {
            NodeList rowNodes = table.getChildNodes();
            for (int i=0;i<rowNodes.getLength();i++)
            {
                Node rowNode = rowNodes.item(i);
                if (rowNode instanceof OdfTableRow)
                {
                    final OdfTableRow row = (OdfTableRow)rowNode;
                    if (row.getStyleName() != null && row.getStyleName().equals("DN_Headers"))
                    {
                        // Skip headers
                        continue;
                    }

                    if (acmd.getIdentityType() == IdentityType.APPLICATION)
                    {
                        results.add(ec.findObjectUsingAID(new Type(clr.classForName(acmd.getFullClassName())),
                            new FieldValues2()
                        {
                            // StateManager calls the fetchFields method
                            public void fetchFields(ObjectProvider sm)
                            {
                                sm.replaceFields(acmd.getAllMemberPositions(), 
                                    new FetchFieldManager(sm, row));
                            }
                            public void fetchNonLoadedFields(ObjectProvider sm)
                            {
                                sm.replaceNonLoadedFields(acmd.getAllMemberPositions(), 
                                    new FetchFieldManager(sm, row));
                            }
                            public FetchPlan getFetchPlanForLoading()
                            {
                                return null;
                            }
                        }, ignoreCache, true));
                    }
                    else if (acmd.getIdentityType() == IdentityType.DATASTORE)
                    {
                        int idIndex = ODFUtils.getColumnIndexForFieldOfClass(acmd, -1);
                        OdfTableCell idCell = (OdfTableCell)row.getChildNodes().item(idIndex);
                        Object idKey = null;
                        if (idCell.getValueType() == OdfValueType.STRING)
                        {
                            idKey = idCell.getStringValue();
                        }
                        else
                        {
                            idKey = Long.valueOf(idCell.getValue().longValue());
                        }
                        OID oid = OIDFactory.getInstance(ec.getOMFContext(), acmd.getFullClassName(), idKey);
                        // TODO This ignores the fields in this row!
                        results.add(ec.findObject(oid, true, true, null));
                    }
                }
            }
        }
        return results;
    }
}